/*
 * pajs_mixer.c

 *
 *  Created on: 10.07.2015
 *      Author: klausj
 */
#include "ips_audio_pulse_PulseMixerProvider.h"
#include "ips_audio_pulse_PulseSourceDataline.h"

#define INITIAL_NATIVE_BUFSIZE 4096

JNIEXPORT jobject JNICALL Java_ips_audio_pulse_PulseSourceDataLine_init(JNIEnv *env, jobject jobj,jobject mpNp) {
	PulseMixerProvider* pmp = (*env)->GetDirectBufferAddress(env,mpNp);
	jobject np = NULL;

	JavaVM *javaVM;
	jint res = (*env)->GetJavaVM(env, &javaVM);
	if (res == 0) {

		PulseSourceDataLine *psdl=(PulseSourceDataLine*)malloc(sizeof(PulseSourceDataLine));
		psdl->paCtx=pmp->paCtx;
		psdl->m=pmp->paMainloop;
		psdl->nativeBuf=NULL;
		psdl->nativeBufLen=0;
		psdl->javaVM = javaVM;
		psdl->javaObjRef = (*env)->NewWeakGlobalRef(env, jobj);

		np= (*env)->NewDirectByteBuffer(env, psdl,sizeof(PulseSourceDataLine));
		jclass cls = (*env)->GetObjectClass(env, psdl->javaObjRef);
		if (cls == NULL) {
			fprintf(stderr,"Could not find class\n");
			return NULL;
		}
		psdl->jClassRef=(*env)->NewGlobalRef(env,cls);
	}
	return(np);
}

// Java : int sampleRate, int validSampleSizeInBits,int sampleSizeInBits, int channels, boolean bigEndian
JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseSourceDataLine_open(JNIEnv *env, jobject obj,jobject np,jstring jsinkName,jint sampleRate, jint validSampleSizeInBits,jint sampleSizeInBits, jint channels, jboolean bigEndian){

	const char *streamName="PulseAudio JavaSound";
	pa_sample_spec  	ss;
	pa_sample_spec_init(&ss);

	ss.channels=channels;
	ss.format=PA_SAMPLE_INVALID;
	if(validSampleSizeInBits==8){
		ss.format=PA_SAMPLE_U8;
	}else if(validSampleSizeInBits==16){

		if(bigEndian){
			ss.format=PA_SAMPLE_S16BE;
		}else{
			ss.format=PA_SAMPLE_S16LE;
		}

	}else if(validSampleSizeInBits==24){
		if(sampleSizeInBits==24){

			if(bigEndian){
				ss.format=PA_SAMPLE_S24BE;
			}else{
				ss.format=PA_SAMPLE_S24LE;
			}
		}
		if(sampleSizeInBits==32){
			if(bigEndian){
				ss.format=PA_SAMPLE_S24_32BE;
			}else{
				ss.format=PA_SAMPLE_S24_32LE;
			}
		}
	}else if(validSampleSizeInBits==32){
		if(bigEndian){
			ss.format=PA_SAMPLE_S32BE;
		}else{
			ss.format=PA_SAMPLE_S32LE;
		}
	}
	if(ss.format==PA_SAMPLE_INVALID){
		fprintf(stderr,"PA: playback sample format invalid\n");
		fflush(stderr);
		(*env)->ThrowNew(env,(*env)->FindClass(env,"ips/audio/pulse/PulseException"),"Playback sample format invalid");
		return;
	}
#ifdef DEBUG
	printf("PA: playback sample format %s\n",pa_sample_format_to_string(ss.format));
	fflush(stdout);
#endif
	ss.rate=sampleRate;

	int res=pa_sample_spec_valid(&ss);
	if(res<=0){
		fprintf(stderr,"PA: sample spec invalid res %d\n",res);
		fflush(stderr);
		(*env)->ThrowNew(env,(*env)->FindClass(env,"ips/audio/pulse/PulseException"),pa_strerror(res));
		return;
	}
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);

	psdl->nativeBuf=malloc(INITIAL_NATIVE_BUFSIZE);
	psdl->nativeBufLen=INITIAL_NATIVE_BUFSIZE;
	psdl->paStream=pa_stream_new (psdl->paCtx,streamName,&ss,NULL);
	pa_stream_set_state_callback(psdl->paStream,play_stream_cb,psdl);

	jboolean isCopy;
	const char *sinkName = (*env)->GetStringUTFChars(env,jsinkName, &isCopy);

//	pa_buffer_attr buffAttr;
//	buffAttr.fragsize=-1;
//	buffAttr.maxlength=-1;
//	buffAttr.minreq=-1;
//	buffAttr.tlength=-1;
//	buffAttr.prebuf=-1;

	//pa_stream_flags_t flags=PA_STREAM_INTERPOLATE_TIMING | PA_STREAM_AUTO_TIMING_UPDATE;
	pa_stream_flags_t flags=PA_STREAM_INTERPOLATE_TIMING | PA_STREAM_AUTO_TIMING_UPDATE | PA_STREAM_START_CORKED;

	pa_threaded_mainloop_lock(psdl->m);
	res=pa_stream_connect_playback 	(psdl->paStream,sinkName,
			NULL,
			flags,
			NULL,
			NULL
	);
	if(res!=0){
		fprintf(stderr,"PA: playback connect: %s\n",pa_strerror(res));
		fflush(stderr);
		(*env)->ThrowNew(env,(*env)->FindClass(env,"ips/audio/pulse/PulseException"),pa_strerror(res));
		return;
	}
	pa_threaded_mainloop_unlock(psdl->m);

	(*env)->ReleaseStringUTFChars(env,jsinkName,sinkName);

}

JNIEXPORT jint JNICALL Java_ips_audio_pulse_PulseSourceDataLine_available(JNIEnv *env, jobject obj,jobject np) {
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);
#ifdef DEBUG
	printf("PA: avail...\n");
#endif
	pa_stream* stream=psdl->paStream;
	//pa_threaded_mainloop_lock(psdl->m);
	jint av=pa_stream_writable_size(stream);
	//pa_threaded_mainloop_unlock(psdl->m);
#ifdef DEBUG
	printf("PA: avail %d bytes\n",av);
	fflush(stdout);
#endif
	return av;
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseSourceDataLine_write(JNIEnv *env, jobject obj,jobject np,jbyteArray jdata,jint offset,jint len) {
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream* stream=psdl->paStream;
#ifdef DEBUG
	printf("PA: write %d bytes...\n",len);
	fflush(stdout);
#endif
	if(len>psdl->nativeBufLen){
		// dynamically increase buffer
		if(psdl->nativeBuf!=NULL){
			free(psdl->nativeBuf);
		}
		psdl->nativeBuf=malloc(len);
		psdl->nativeBufLen=len;
	}
	(*env)->GetByteArrayRegion(env,jdata,offset,len,psdl->nativeBuf);
	pa_threaded_mainloop_lock(psdl->m);
	int res=pa_stream_write(stream,psdl->nativeBuf,len,NULL,0,PA_SEEK_RELATIVE);
	if(res){
		fprintf(stderr,"PA: write res %d %s\n",res,pa_strerror(res));
		fflush(stderr);
		(*env)->ThrowNew(env,(*env)->FindClass(env,"ips/audio/pulse/PulseException"),pa_strerror(res));
		pa_threaded_mainloop_unlock(psdl->m);
		return;
	}
	pa_threaded_mainloop_unlock(psdl->m);
#ifdef DEBUG
	printf("PA: write %d bytes\n",len);
	fflush(stdout);
#endif
}


// TODO Move cork and flush code to common PulseDataline

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseSourceDataLine_cork(JNIEnv *env, jobject obj,jobject np,jboolean cork) {
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream* stream=psdl->paStream;

	pa_threaded_mainloop_lock(psdl->m);
	int b=cork?1:0;
	int corked=pa_stream_is_corked(stream);
	if(b!=corked){
		pa_operation *op=pa_stream_cork(stream,b,cork_cb,psdl);
		if(op){
			pa_operation_unref(op);
		}else{
			fprintf(stderr,"Cork pa_operation == NULL !\n");
			fflush(stderr);
		}
	}else{
#ifdef DEBUG
		// TODO why is this happening sometimes
		fprintf(stderr,"Stream already in corked state %d!\n",corked);
		fflush(stderr);
#endif
	}

	pa_threaded_mainloop_unlock(psdl->m);

}

void play_stream_cb(pa_stream *s,void *userdata){

	PulseSourceDataLine* psdl = userdata;
	pa_stream_state_t sst=pa_stream_get_state 	(psdl->paStream);
	char* mNm=NULL;
	if(sst==PA_STREAM_FAILED){
		mNm="streamFailed";
	}else if(sst==PA_STREAM_READY){
		mNm="streamReady";
	}if(sst==PA_STREAM_TERMINATED){
		mNm="streamTerminated";
	}
	if(mNm!=NULL){
		JNIEnv *env;
		JavaVM *jvm = psdl->javaVM;
		if (jvm != NULL) {

			jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
			if (res == 0) {
				jclass cls = (*env)->GetObjectClass(env, psdl->javaObjRef);
				if (cls == NULL) {
					printf("Could not find class\n");
					return;
				}
				jmethodID mId=NULL;

				mId=(*env)->GetMethodID(env,cls,mNm,"()V");
				if(mId==NULL){
					fprintf(stderr,"Method not found!");
					return;
				}
				(*env)->CallVoidMethod(env,psdl->javaObjRef,mId);
				(*jvm)->DetachCurrentThread(jvm);
			}
		}
	}
}

void cork_cb(pa_stream *s, int success, void *userdata){


	PulseSourceDataLine* psdl = userdata;
	JNIEnv *env;
	JavaVM *jvm = psdl->javaVM;
	if (jvm != NULL) {

		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res == 0) {
			jclass cls = (*env)->GetObjectClass(env, psdl->javaObjRef);
			if (cls == NULL) {
				fprintf(stderr,"Could not find class\n");
				return;
			}
			jmethodID mId=NULL;

			mId=(*env)->GetMethodID(env,cls,"corkStatusChanged","(I)V");
			if(mId==NULL){
				fprintf(stderr,"Method not found!");
				return;
			}
			(*env)->CallVoidMethod(env,psdl->javaObjRef,mId,success);
			(*jvm)->DetachCurrentThread(jvm);
		}
	}
}

void flush_cb(pa_stream *s, int success, void *userdata){
#ifdef DEBUG
	printf("PA: flush callback\n");
    fflush(stdout);
#endif
	PulseSourceDataLine* psdl = userdata;
			JNIEnv *env;
			JavaVM *jvm = psdl->javaVM;
			if (jvm != NULL) {

				jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
				if (res == 0) {
					jclass cls = (*env)->GetObjectClass(env, psdl->javaObjRef);
					if (cls == NULL) {
						printf("Could not find class\n");
						return;
					}
					jmethodID mId=NULL;

					mId=(*env)->GetMethodID(env,cls,"flushed","(I)V");
					if(mId==NULL){
						fprintf(stderr,"Method not found!");
						return;
					}
					(*env)->CallVoidMethod(env,psdl->javaObjRef,mId,success);
					(*jvm)->DetachCurrentThread(jvm);
				}
			}
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseSourceDataLine_trigger(JNIEnv *env, jobject obj,jobject np) {
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream* stream=psdl->paStream;
	pa_threaded_mainloop_lock(psdl->m);
	pa_operation *op=pa_stream_trigger(stream,NULL,psdl);
	pa_operation_unref(op);
	pa_threaded_mainloop_unlock(psdl->m);
#ifdef DEBUG
	printf("PA: trigger\n");
	fflush(stdout);
#endif
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseSourceDataLine_flushBegin(JNIEnv *env, jobject obj,jobject np) {
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream* stream=psdl->paStream;
	pa_threaded_mainloop_lock(psdl->m);
	pa_operation *op=pa_stream_flush(stream,flush_cb,psdl);
	pa_operation_unref(op);
	pa_threaded_mainloop_unlock(psdl->m);
#ifdef DEBUG
	printf("PA: flush\n");
	fflush(stdout);
#endif

}

JNIEXPORT jlong JNICALL Java_ips_audio_pulse_PulseSourceDataLine_bytePosition(JNIEnv *env, jobject obj,jobject np) {
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);

	pa_stream* stream=psdl->paStream;
	jlong bytePos=0;
	pa_threaded_mainloop_lock(psdl->m);
	// TODO not asynchronous
	pa_stream_update_timing_info(stream,NULL,psdl);
	const pa_timing_info* ti=pa_stream_get_timing_info(stream);
	if(ti!=NULL){
		bytePos=ti->read_index;
	}
	pa_threaded_mainloop_unlock(psdl->m);

	return bytePos;
}

JNIEXPORT jlong JNICALL Java_ips_audio_pulse_PulseSourceDataLine_usecPosition(JNIEnv *env, jobject obj,jobject np) {
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);

	pa_stream* stream=psdl->paStream;
	pa_usec_t usecPos=0;

	int r=pa_stream_get_time(stream,&usecPos);

	if(!r){
		// Hmmm, returns 0
	}

	return usecPos;
}


void drain_cb(pa_stream *s, int success, void *userdata){
#ifdef DEBUG
	printf("PA: drained1 %d\n",success);
	fflush(stdout);
#endif
	PulseSourceDataLine* psdl = userdata;
	JNIEnv *env;
	JavaVM *jvm = psdl->javaVM;
	if (jvm != NULL) {

		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res == 0) {
			jclass cls = (*env)->GetObjectClass(env, psdl->javaObjRef);
			if (cls == NULL) {
				fprintf(stderr,"Could not find class\n");
				return;
			}
			jmethodID mId=NULL;

			mId=(*env)->GetMethodID(env,cls,"drained","(I)V");
			if(mId==NULL){
				fprintf(stderr,"Method not found!");
				return;
			}
			(*env)->CallVoidMethod(env,psdl->javaObjRef,mId,success);
			(*jvm)->DetachCurrentThread(jvm);
		}
	}
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseSourceDataLine_drainBegin(JNIEnv *env, jobject obj,jobject np) {
	PulseSourceDataLine* psdl = (*env)->GetDirectBufferAddress(env,np);
	pa_stream* stream=psdl->paStream;
	pa_threaded_mainloop_lock(psdl->m);
	pa_operation *op=pa_stream_drain(stream,drain_cb,psdl);
	pa_operation_unref(op);
	pa_threaded_mainloop_unlock(psdl->m);
#ifdef DEBUG
	printf("PA: drain\n");
	fflush(stdout);
#endif
}




JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseSourceDataLine_close(JNIEnv *env, jobject obj,jobject np) {
	PulseSourceDataLine *psdl = (*env)->GetDirectBufferAddress(env,np);
	pa_threaded_mainloop_lock(psdl->m);
	int res=pa_stream_disconnect(psdl->paStream);
	pa_threaded_mainloop_unlock(psdl->m);
	if(res){
		fprintf(stderr,"PA: play stream disconnect: %s\n",pa_strerror(res));
		fflush(stderr);
	}else{
#ifdef DEBUG
		printf("PA: play stream disconnect\n");
		fflush(stdout);
#endif
	}
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseSourceDataLine_release(JNIEnv *env, jobject obj,jobject np) {
	PulseSourceDataLine *psdl = (*env)->GetDirectBufferAddress(env,np);
	if(psdl->paStream){
		pa_stream_unref(psdl->paStream);
	}
	psdl->nativeBufLen=0;
	if(psdl->nativeBuf!=NULL){
		free(psdl->nativeBuf);
	}
}


